module integrator
    real(kind=8), parameter :: IDLETIME = 0.1
    integer, parameter :: TAG_WORK = 0
    integer, parameter :: TAG_END = 2
contains

    function func(x)
        use mpi
        implicit none
        real(kind=8) :: func
        real(kind=8), intent(in) :: x

        real(kind=8), parameter :: IDLETIME = 0.1
        real(kind=8) :: t

        t = MPI_WTIME()
        ! Introduce work imballance by sleeping more at larger x
        do while (MPI_WTIME()-t <= IDLETIME*x*x)
        end do
        func = 4.0_8 / (1.0_8 + x*x)
    end function func

    function controller(x_start, x_end, maxSteps, rangeSize) result(sum)
        use mpi
        implicit none
        real(kind=8), intent(in) :: x_start, x_end
        integer, intent(in) :: maxSteps, rangeSize

        integer :: myRank, numProcs, ierr
        real(kind=8) :: res, sum, stepSize
        integer, dimension(2) :: s
        integer :: range, maxRanges, nextRank, items
        integer, dimension(MPI_STATUS_SIZE) :: status

        call MPI_COMM_SIZE(MPI_COMM_WORLD, numProcs, ierr)

        sum = 0.0

        maxRanges = ceiling(real(maxSteps,8)/rangeSize)

        ! Part 1: Send work to everyone
        range = 0
        do while (range < numProcs-1 .and. range < maxRanges)
            s(1) = rangeSize*range
            s(2) = min(rangeSize*(range+1),maxSteps)
            nextRank = range+1
            call MPI_SEND(s, 2, MPI_INTEGER, nextRank, TAG_WORK, &
                MPI_COMM_WORLD, ierr)
            range = range + 1
        end do

        ! Part 2: Receive 'maxRanges' result items from our workers
        items = 0
        do while (items < maxRanges)
            ! Receive a work item from anyone
            call MPI_RECV(res, 1, MPI_REAL8, MPI_ANY_SOURCE, TAG_WORK, &
                MPI_COMM_WORLD, status, ierr)
            sum = sum + res
            ! Next work item goes to the same worker - it is now fre
            nextRank = status(MPI_SOURCE)

            ! Send some work if there is still any left
            if (range < maxRanges) then
                s(1) = rangeSize*range
                s(2) = min(rangeSize*(range+1),maxSteps)
                call MPI_SEND(s, 2, MPI_INTEGER, nextRank, TAG_WORK, &
                    MPI_COMM_WORLD, ierr)
                range = range + 1
            end if

            items = items + 1
        end do

        ! Signal workers to stop by sending empty messages with tag TAG_END
        do nextRank = 1, numProcs-1
            call MPI_SEND(nextRank, 0, MPI_INTEGER, nextRank, TAG_END, &
                MPI_COMM_WORLD, ierr)
        end do
    end function controller

    subroutine worker(f, x_start, x_end, maxSteps)
        use mpi
        implicit none
        interface
            function f(x)
                real(kind=8) :: f
                real(kind=8), intent(in) :: x
            end function f
        end interface
        real(kind=8), intent(in) :: x_start, x_end
        integer, intent(in) :: maxSteps

        integer, dimension(2) :: s
        real(kind=8), dimension(2) :: x, y
        integer :: ierr, step
        real(kind=8) :: sum, stepSize
        integer, dimension(MPI_STATUS_SIZE) :: status

        stepSize = (x_end - x_start) / maxSteps

        work_loop: do while(.true.)
            ! I am worker, wait for work

            ! Receive the left and right points of the trapezoid and compute
            ! the corresponding function values. If the tag is TAG_END, don't
            ! compute but exit.
            call MPI_RECV(s, 2, MPI_INTEGER, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &
                status, ierr)
            if (status(MPI_TAG) == TAG_END) exit work_loop

            ! Integrate from step s(1) to step s(2)
            sum = 0.0
            do step = s(1), s(2)-1
                x(1) = x_start + stepSize*step
                x(2) = x_start + stepSize*(step+1)
                y(1) = f(x(1))
                y(2) = f(x(2))
                sum = sum + stepSize*0.5*(y(1)+y(2))
            end do
            ! Send back the computed result
            call MPI_SEND(sum, 1, MPI_REAL8, 0, TAG_WORK, MPI_COMM_WORLD, ierr)
        end do work_loop
    end subroutine worker

    function integrate(f, x_start, x_end, maxSteps, rangeSize) result(sum)
        use mpi
        implicit none
        interface
            function f(x)
                real(kind=8) :: f
                real(kind=8), intent(in) :: x
            end function f
        end interface
        real(kind=8), intent(in) :: x_start, x_end
        integer, intent(in) :: maxSteps, rangeSize

        real(kind=8) :: sum
        integer :: myRank, ierr

        call MPI_COMM_RANK(MPI_COMM_WORLD, myRank, ierr)

        if (myRank == 0) then
            sum = controller(x_start, x_end, maxSteps, rangeSize)
        else
            call worker(f, x_start, x_end, maxSteps)
            sum = 0.0
        end if
    end function integrate

end module integrator

program calc_pi
    use mpi
    use ifport
    use integrator
    implicit none
    integer :: myRank, numProcs, ierr
    integer :: maxSteps = 100
    integer :: rangeSize = 10
    real(kind=8) :: startTime, stopTime, pi
    character(len=128) :: cmdArg
    ! Integration domain si [0, 1]
    real(kind=8), parameter :: x0 = 0.0, x1 = 1.0

    call MPI_INIT(ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, myRank, ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, numProcs, ierr)

    if (iargc() > 0) then
        call getarg(1, cmdArg)
        read(cmdArg,*) maxSteps
        if (maxSteps < 1) call MPI_ABORT(MPI_COMM_WORLD, 1, ierr)
    end if
    if (iargc() > 1) then
        call getarg(2, cmdArg)
        read(cmdArg,*) rangeSize
        if (rangeSize < 1) call MPI_ABORT(MPI_COMM_WORLD, 1, ierr)
    end if

    if (myRank == 0) then
        print '("Integrating from ",F9.6," to ",F9.6," in ",I0," steps &
            &(range size ",I0,")")', &
            x0, x1, maxSteps, rangeSize
    end if

    ! Synchronize before making performance measurements
    call MPI_BARRIER(MPI_COMM_WORLD, ierr)

    startTime = MPI_WTIME()

    pi = integrate(func, x0, x1, maxSteps, rangeSize)

    stopTime = MPI_WTIME()

    if (myRank == 0) then
        print '(/,"PI = ",F9.6,"    integral = ",F9.6)', 4.0_8*atan(1.0_8), pi
        print '("Computation took ",F10.6," seconds")', stopTime-startTime
    end if

    call MPI_FINALIZE(ierr)
end program calc_pi
