!
! Program to solve a finite difference
! discretization of the screened Poisson equation:
! (d2/dx2)u + (d2/dy2)u - alpha u = f
! with zero Dirichlet boundary condition using the iterative
! Jacobi method with overrelaxation.
!
! RHS (source) function
!   f(x,y) = -alpha*(1-x^2)(1-y^2)-2*[(1-x^2)+(1-y^2)]
!
! Analytical solution to the PDE
!   u(x,y) = (1-x^2)(1-y^2)
!
! Current Version: Christian Iwainsky, RWTH Aachen University
! MPI C Version: Christian Terboven, RWTH Aachen University, 2006
! MPI Fortran Version: Dieter an Mey, RWTH Aachen University, 1999 - 2005
! Modified: Sanjiv Shah,        Kuck and Associates, Inc. (KAI), 1998
! Author:   Joseph Robicheaux,  Kuck and Associates, Inc. (KAI), 1998
!
! Unless READ_INPUT is defined, a meaningful input dataset is used (CT).
!
! Input : n     - grid dimension in x direction
!         m     - grid dimension in y direction
!         alpha - constant (always greater than 0.0)
!         tol   - error tolerance for the iterative solver
!         relax - Successice Overrelaxation parameter
!         mits  - maximum iterations for the iterative solver
!
! On output
!       : u(n,m)       - Dependent variable (solution)
!       : f(n,m,alpha) - Right hand side function
!

module solver
contains

    !
    ! Performs one iteraton of the Jacobi method and computes
    ! the residual value.
    !
    ! NOTE: u(0,:), u(maxXCount,:), u(:,0) and u(:,maxYCount)
    ! are BOUNDARIES and therefore not part of the solution.
    !
    subroutine one_jacobi_iteration(xStart, yStart, maxXCount, maxYCount, &
        src, dst, deltaX, deltaY, alpha, omega, error)
        implicit none
        real(kind=8), intent(in) :: xStart, yStart
        integer, intent(in) :: maxXCount, maxYCount
        real(kind=8), dimension(0:maxXCount-1,0:maxYCount-1), intent(in) :: src
        real(kind=8), dimension(0:maxXCount-1,0:maxYCount-1), intent(out) :: dst
        real(kind=8), intent(in) :: deltaX, deltaY, alpha, omega
        real(kind=8), intent(out) :: error

        integer :: x, y
        real(kind=8) :: f, fX, fY
        real(kind=8) :: updateVal
        ! Coefficients
        real(kind=8) :: cx, cy, cc
        cx = 1.0_8/(deltaX**2)
        cy = 1.0_8/(deltaY**2)
        cc = -2.0_8*cx-2.0_8*cy-alpha

        error = 0.0

        do y = 1, maxYCount-2
            fY = yStart + (y-1)*deltaY
            do x = 1, maxXCount-2
                fX = xStart + (x-1)*deltaX
                f = -alpha*(1.0-fX*fX)*(1.0-fY*fY)-2.0*(1.0-fX*fX)-2.0*(1.0-fY*fY)
                updateVal = ((src(x-1,y)+src(x+1,y))*cx + &
                    (src(x,y-1)+src(x,y+1))*cy + &
                    src(x,y)*cc - f)/cc
                dst(x,y) = src(x,y) - omega*updateVal
                error = error + updateVal*updateVal
            end do
        end do
        error = sqrt(error)/((maxXCount-2)*(maxYCount-2))
    end subroutine one_jacobi_iteration

    !
    ! Checks the error between numerical and exact solutions
    !
    subroutine checkSolution(xStart, yStart, maxXCount, maxYCount, u, &
        deltaX, deltaY, alpha, error)
        implicit none
        real(kind=8), intent(in) :: xStart, yStart
        integer, intent(in) :: maxXCount, maxYCount
        real(kind=8), dimension(0:maxXCount-1,0:maxYCount-1), intent(in) :: u
        real(kind=8), intent(in) :: deltaX, deltaY, alpha
        real(kind=8), intent(out) :: error

        integer :: x, y
        real(kind=8) :: fX, fY
        real(kind=8) :: localError

        error = 0.0

        do y = 1, maxYCount-2
            fY = yStart + (y-1)*deltaY
            do x = 1, maxXCount-2
                fX = xStart + (x-1)*deltaX
                localError = u(x,y) - (1.0-fX**2)*(1.0-fY**2)
                error = error + localError**2
            end do
        end do
        error = sqrt(error)/((maxXCount-2)*(maxYCount-2))
    end subroutine checkSolution

end module solver

!
! Main program
!
program jacobi
    use solver
    implicit none
    integer :: n, m, mits
    real(kind=8) :: alpha, tol, relax
    real(kind=8) :: maxAcceptableError
    real(kind=8) :: error, absoluteError
    real(kind=8), dimension(:,:), allocatable, target :: u1, u2
    real(kind=8), dimension(:,:), pointer :: u, u_old, tmp
    integer :: iterationCount, maxIterationCount

    ! Solution space is [-1, 1] x [-1, 1]
    real(kind=8), parameter :: xLeft = -1.0, xRight = 1.0
    real(kind=8), parameter :: yBottom = -1.0, yUp = 1.0

    real(kind=8) :: deltaX, deltaY

#ifdef READ_INPUT
    print *,"Input n m - grid dimension in x and y direction:"
    read *,n,m
    print *,"Input alpha - Helmholtz constant:"
    read *,alpha
    print *,"Input relax - successive over-relaxation parameter:"
    read *,relax
    print *,"Input tol - error tolerance for the iterative solver:"
    read *,tol
    print *,"Input mits - maximum solver iterations:"
    read *,mits
#else
    n = 200
    m = 200
    alpha = 0.8
    relax = 1.0
    tol = 1.e-7
    mits = 1000
#endif
    print *,"->",n,m,alpha,relax,tol,mits

    allocate(u1(0:n+1,0:m+1))
    u_old => u1
    allocate(u2(0:n+1,0:m+1))
    u => u2

    ! Zero the elements (including the boundaries)
    u_old = 0.
    u = 0.

    maxIterationCount = mits
    maxAcceptableError = tol

    deltaX = (xRight-xLeft)/(n-1)
    deltaY = (yUp-yBottom)/(m-1)

    iterationCount = 0
    error = 1.e+100_8

    ! Iterate as long as it takes to meet the convergence criterion
    do while (iterationCount < maxIterationCount .and. &
        error > maxAcceptableError)
        print *,"Iteration",iterationCount
        call one_jacobi_iteration(xLeft, yBottom, n+2, m+2, u_old, u, &
            deltaX, deltaY, alpha, relax, error)
        print *,"   Error",error
        iterationCount = iterationCount+1
        ! Swap the buffers
        tmp => u_old
        u_old => u
        u => tmp
    end do
    print *,"Residual",error

    ! u_old points to the solution after the most recent buffers swap
    call checkSolution(xLeft, yBottom, n+2, m+2, u_old, &
        deltaX, deltaY, alpha, absoluteError)
    print *,"The error of the iterative solution is", absoluteError
end program jacobi
